@@ -3,9 +3,11 @@ from __future__ import annotations
 import typing
 from abc import ABC, abstractmethod
 from collections import OrderedDict
+from collections.abc import Callable, Iterator
 from inspect import isclass
 from pathlib import Path
-from typing import Any, Callable, Dict, Generic, Iterator, List, Literal, Optional, Set, TypeVar, Union, cast
+from types import UnionType
+from typing import Any, Generic, Literal, TypeVar, Union, cast
 
 from tox.config.types import Command, EnvList
 
@@ -13,13 +15,13 @@ _NO_MAPPING = object()
 T = TypeVar("T")
 V = TypeVar("V")
 
-Factory = Optional[Callable[[object], T]]  # note the argument is anything, due e.g. memory loader can inject anything
+Factory = Callable[[object], T] | None  # note the argument is anything, due e.g. memory loader can inject anything
 
 
 class Convert(ABC, Generic[T]):
     """A class that converts a raw type to a given tox (python) type."""
 
-    def to(self, raw: T, of_type: type[V], factory: Factory[V]) -> V:  # noqa: PLR0911, C901
+    def to(self, raw: T, of_type: type[V] | UnionType, factory: Factory[V]) -> V:  # noqa: PLR0911
         """
         Convert given raw type to python type.
 
@@ -29,7 +31,11 @@ class Convert(ABC, Generic[T]):
         :return: the converted type
         """
         from_module = getattr(of_type, "__module__", None)
-        if from_module in {"typing", "typing_extensions"}:
+        if (
+            from_module in {"typing", "typing_extensions"}
+            or of_type.__class__ == UnionType
+            or (hasattr(typing, "GenericAlias") and isinstance(of_type, typing.GenericAlias))
+        ):
             return self._to_typing(raw, of_type, factory)
         if isclass(of_type):
             if issubclass(of_type, Path):
@@ -42,35 +48,32 @@ class Convert(ABC, Generic[T]):
                 return self.to_env_list(raw)  # type: ignore[return-value]
             if issubclass(of_type, str):
                 return self.to_str(raw)  # type: ignore[return-value]
-        # python does not allow use of parametrized generics with isinstance, so we need to check for them.
-        if hasattr(typing, "GenericAlias") and isinstance(of_type, typing.GenericAlias):
-            return list(self.to_list(raw, of_type=of_type))  # type: ignore[return-value]
-        if isinstance(raw, of_type):  # already target type no need to transform it
+        if isinstance(raw, cast("type[V]", of_type)):  # already target type no need to transform it
             # do it this late to allow normalization - e.g. string strip
-            return raw  # type: ignore[no-any-return]
+            return raw
         if factory:
             return factory(raw)
-        return of_type(raw)  # type: ignore[no-any-return]
+        return cast("type[V]", of_type)(raw)  # type: ignore[call-arg]
 
-    def _to_typing(self, raw: T, of_type: type[V], factory: Factory[V]) -> V:  # noqa: C901
+    def _to_typing(self, raw: T, of_type: type[V] | UnionType, factory: Factory[V]) -> V:  # noqa: C901
         origin = getattr(of_type, "__origin__", of_type.__class__)
         result: Any = _NO_MAPPING
-        if origin in {list, List}:
-            entry_type = of_type.__args__[0]  # type: ignore[attr-defined]
+        if origin in {list, list}:
+            entry_type = cast("type[V]", of_type).__args__[0]  # type: ignore[attr-defined]
             result = [self.to(i, entry_type, factory) for i in self.to_list(raw, entry_type)]
             if isclass(entry_type) and issubclass(entry_type, Command):
                 result = [i for i in result if i is not None]
-        elif origin in {set, Set}:
-            entry_type = of_type.__args__[0]  # type: ignore[attr-defined]
+        elif origin in {set, set}:
+            entry_type = cast("type[V]", of_type).__args__[0]  # type: ignore[attr-defined]
             result = {self.to(i, entry_type, factory) for i in self.to_set(raw, entry_type)}
-        elif origin in {dict, Dict}:
-            key_type, value_type = of_type.__args__[0], of_type.__args__[1]  # type: ignore[attr-defined]
+        elif origin in {dict, dict}:
+            key_type, value_type = cast("type[V]", of_type).__args__[0], cast("type[V]", of_type).__args__[1]  # type: ignore[attr-defined]
             result = OrderedDict(
                 (self.to(k, key_type, factory), self.to(v, value_type, factory))
                 for k, v in self.to_dict(raw, (key_type, value_type))
             )
-        elif origin == Union:  # handle Optional values
-            args: list[type[Any]] = of_type.__args__  # type: ignore[attr-defined]
+        elif origin in {Union, UnionType}:  # handle Optional values
+            args: list[type[Any]] = of_type.__args__  # type: ignore[union-attr,assignment]
             none = type(None)
             if len(args) == 2 and none in args:  # noqa: PLR2004
                 if isinstance(raw, str):
@@ -81,13 +84,13 @@ class Convert(ABC, Generic[T]):
                     new_type = next(i for i in args if i != none)  # pragma: no cover
                     result = self.to(raw, new_type, factory)
         elif origin in {Literal, type(Literal)}:
-            choice = of_type.__args__  # type: ignore[attr-defined]
+            choice = cast("type[V]", of_type).__args__  # type: ignore[attr-defined]
             if raw not in choice:
                 msg = f"{raw} must be one of {choice}"
                 raise ValueError(msg)
             result = raw
         if result is not _NO_MAPPING:
-            return cast(V, result)
+            return cast("V", result)
         msg = f"{raw} cannot cast to {of_type!r}"
         raise TypeError(msg)
 
